#include "stdafx.h"
#include "MainForm.h"
#include "MachineState.h"
#include "Disassembly.h"
#include "MicroDisassembly.h"
#include "InputDialog.h"

using namespace Simulator;
using namespace System::Windows::Forms;
using namespace System::IO;

// Initialize the simulator GUI
void MainForm::InitializeSimulation()
{
	m_tooltip->SetToolTip(m_continueButtton, "Run");
	m_tooltip->SetToolTip(m_stepButton, "Step Over");
	m_tooltip->SetToolTip(m_stepInButton, "Step In");
	m_tooltip->SetToolTip(m_stepOutButton, "Step Out");
	m_tooltip->SetToolTip(m_microStepButton, "Micro-Step");
	m_tooltip->SetToolTip(m_pauseButton, "Pause");
	m_tooltip->SetToolTip(m_resetButton, "Reset");

	try
	{
		String^ loadAddress = safe_cast<String^>(Application::UserAppDataRegistry->GetValue("Load Address"));
		m_loadAddress = UInt32::Parse(loadAddress);
	}
	catch(Exception^)
	{
		m_loadAddress = 0x010002;
	}

	try
	{
		m_bootImage = safe_cast<String^>(Application::UserAppDataRegistry->GetValue("Boot Image"));
		if (!m_bootImage)
		{
			m_bootImage = ".\\microchess.bin";
		}
	}
	catch(Exception^)
	{
		m_bootImage = ".\\microchess.bin";
	}

	try
	{
		String^ bootAddress = safe_cast<String^>(Application::UserAppDataRegistry->GetValue("Boot Address"));
		m_bootAddress = UInt32::Parse(bootAddress);
	}
	catch(Exception^)
	{
		m_bootAddress = 0x010002;
	}

	try
	{
		m_refreshCycles = safe_cast<unsigned long>(Application::UserAppDataRegistry->GetValue("Refresh Cycles"));
	}
	catch(Exception^)
	{
		m_refreshCycles = 200000;
	}

	m_pauseButton->Enabled = false;
	m_running = false;	
	m_stepsSinceLastInterrupt = 0;
	m_memoryViewAddress = 0;
	m_sourceDisassemblyStartAddress = 0xFFFFFFFF;
	m_memoryBreakpointAddress = 0xFFFFFF;
	m_machineState = gcnew MachineState(m_bootImage, ".", m_loadAddress, m_bootAddress);
	m_disassembler = gcnew Disassembly(m_machineState);
	m_microDisassembler = gcnew MicroDisassembly(m_machineState);

	m_currentInstructionAddress = m_machineState->RegPC - 1;
	m_currentStackFrameAddress = m_currentInstructionAddress;

	// Set the initial condition codes so all the lights are off. In the real
	// hardware, they are uninitialized at startup.
	m_machineState->FlagNotAddressCarry = true;
	m_machineState->FlagNotCarry = true;
	m_machineState->FlagNotZero = true;
	m_machineState->FlagNotOverflow = true;

	array<String^>^ parts = m_bootImage->Split('\\');
	this->Text = "BMOW - " + parts[parts->Length - 1];
	
	LoadSourceLabels(m_bootImage->Substring(0, m_bootImage->LastIndexOf('.')) + ".sym");

	m_stepOverDestinationAddress = 0;
	m_steppingOut = false;
	m_stepOutCount = 0;

	m_stackFrames.Clear();

	RefreshData();
}

bool MainForm::RetireInstruction()
{
	bool forceBreak = false;
	
	if (m_executingOpcode == 0x60 || // RTS
		m_executingOpcode == 0x6B)   // RTL
	{
		if (m_steppingOut && m_stepOutCount == 0)
		{
			m_steppingOut = false;
			forceBreak = true;
		}

		// pop stack
		if (m_stackFrames.Count)
		{
			m_stackFrames.RemoveAt(m_stackFrames.Count - 1);
			m_stepOutCount--;
		}
	}

	if (m_executingOpcode == 0x20 || // JSR
		m_executingOpcode == 0x22)   // JSL
	{
		// push stack
		m_stackFrames.Add(m_executingPC-1);
		m_stepOutCount++;
	}

	// next instruction
	m_executingOpcode = m_machineState->RegOpcode;
	m_executingPC = m_machineState->RegPC;

	return forceBreak;
}

// handle the application idle event, to do background processing
void MainForm::OnIdle(System::Object^  sender, System::EventArgs^  e)
{
	if (m_running)
	{
		TimeSpan frameTime = DateTime::Now - m_lastUpdateTime;
		unsigned int numInstructionsToRun = m_refreshCycles; 
		
		while (numInstructionsToRun)
		{
			bool hitMicroBreakpoint = false;
			bool forceBreak = false;

			do
			{
#define PERIODIC_INTERRUPT
#ifdef PERIODIC_INTERRUPT
				// interrupt?
				if (m_stepsSinceLastInterrupt == 2000)
				{
					m_machineState->FlagNotIrq = false;
					m_stepsSinceLastInterrupt = 0;
				}
				else
				{
					m_machineState->FlagNotIrq = true;
					m_stepsSinceLastInterrupt++;
				}
#endif

				unsigned long addressModified = m_machineState->MicroStep();
				if (addressModified == m_memoryBreakpointAddress)
				{
					MessageBox::Show(String::Format("Memory breakpoint {0:X6}", addressModified));
					forceBreak = true;
					break;
				}

				// microbreakpoint hit?
				if (CurrentMicroaddressIsAtBreakpoint())
				{
					hitMicroBreakpoint = true;
					break;
				}
			} while (m_machineState->Phase != 0);

			// PC points to the next byte after the instruction
			if (m_machineState->Phase == 0)
			{
				m_currentInstructionAddress = m_machineState->RegPC - 1;
				m_currentStackFrameAddress = m_currentInstructionAddress;
				forceBreak = RetireInstruction();
			}

			// breakpoint or halt?
			if (hitMicroBreakpoint ||
				m_sourceBreakpoints.ContainsKey(m_currentInstructionAddress) ||
				m_currentInstructionAddress == m_stepOverDestinationAddress ||
				forceBreak ||
				m_machineState->RegOpcode == 0xFF)
			{
				m_running = false;
				m_consoleTextBox->Focus();
				m_consoleTextBox->SelectionLength = 0;
				m_pauseButton->Enabled = false;
				m_continueButtton->Enabled = true;
				m_stepButton->Enabled = true;
				m_stepInButton->Enabled = true;
				m_stepOutButton->Enabled = true;
				m_microStepButton->Enabled = true;
				m_clearBreakpointsButton->Enabled = true;
				m_stepOverDestinationAddress = 0;
				
				break;
			}

			numInstructionsToRun--;
		}

		// TODO: the simulation actually seems to run more slowly if RefreshData() isn't called
		// I think the Application::Idle event is generated less frequently if the UI doesn't change
		RefreshData();
	}

	m_lastUpdateTime = DateTime::Now;
}

// Test if the current microinstruction address matches a micro-source breakpoint
bool MainForm::CurrentMicroaddressIsAtBreakpoint()
{
	unsigned long microaddress = m_machineState->GetCurrentMicroaddress();
	unsigned long microProgramBaseAddress = microaddress >> 5;
	
	if (m_microSourceBreakpoints.ContainsKey(microProgramBaseAddress))
	{
		List<unsigned long> conditionCodes = m_microSourceBreakpoints[microProgramBaseAddress];
		for each (unsigned long conditionCode in conditionCodes)
		{
			if ((conditionCode & (1 << (microaddress & 0x1))) &&
				(conditionCode & (1 << (((microaddress >> 1) & 0x1) + 2))) &&
				(conditionCode & (1 << (((microaddress >> 2) & 0x1) + 4))) &&
				(conditionCode & (1 << (((microaddress >> 3) & 0x1) + 6))) &&
				(conditionCode & (1 << (((microaddress >> 4) & 0x1) + 8))))
			{
				return true;
			}
		}

		return false;
	}
	else
	{
		return false;
	}
}

// refresh the UI with the machine state data
void MainForm::RefreshData()
{
	// 8-bit registers
	m_aLabel->Text = String::Format("{0:X2}", m_machineState->RegA);
	m_xLabel->Text = String::Format("{0:X2}", m_machineState->RegX);
	m_yLabel->Text = String::Format("{0:X2}", m_machineState->RegY);
	m_tLabel->Text = String::Format("{0:X2}", m_machineState->RegT);

	// address registers
	m_pcLabel->Text = String::Format("{0:X6}", m_machineState->RegPC);
	m_arLabel->Text = String::Format("{0:X6}", m_machineState->RegAR);
	m_spLabel->Text = String::Format("{0:X6}", m_machineState->RegSP);

	// control
	m_opcodeLabel->Text = String::Format("{0:X2}", m_machineState->RegOpcode);
	m_phaseLabel->Text = String::Format("{0:X1}", m_machineState->Phase);
	m_opLabel->Text = m_disassembler->GetOpcodeDescription(m_machineState->RegOpcode);

	// flags
	m_aRadioButton->Checked = !m_machineState->FlagNotAddressCarry;
	m_cRadioButton->Checked = !m_machineState->FlagNotCarry;
	m_nRadioButton->Checked = m_machineState->FlagNegative;
	m_zRadioButton->Checked = !m_machineState->FlagNotZero;
	m_vRadioButton->Checked = !m_machineState->FlagNotOverflow;
	m_ieRadioButton->Checked = !m_machineState->FlagNotInterruptEnable;
	m_irqRadioButton->Checked = !m_machineState->FlagNotIrq;

	// memory
	m_memoryAddressTextBox->Text = String::Format("{0:X6}", m_memoryViewAddress);
	FillMemoryTextBox(m_memoryViewTextBox, m_memoryViewAddress, 0, false);
	FillConsoleTextBox(m_consoleTextBox);

	// call stack
	RefreshCallStack();

	// disassembly
	RefreshDisassemblyView(true);
	RefreshMicroDisassemblyView();

	// LCD
	m_lcdTextBox->Text = m_machineState->LCDText();

	m_microDisassemblyViewTitle->Text = m_disassembler->GetOpcodeDescription(m_machineState->RegOpcode) + " - Micro-Disassembly";
}

// Refresh the source disassembly view
void MainForm::RefreshDisassemblyView(bool scrollToCurrentInstruction)
{
	unsigned long newSourceDisassemblyStartAddress = Math::Max(0,(m_currentStackFrameAddress-50)/50*50);
	if (newSourceDisassemblyStartAddress == m_sourceDisassemblyStartAddress)
	{
		m_disassembler->UpdateDisassemblyListView(m_sourceListView, m_currentInstructionAddress, m_currentStackFrameAddress, %m_sourceBreakpoints, %m_addressToSource, scrollToCurrentInstruction);
	}
	else
	{
		m_sourceDisassemblyStartAddress = newSourceDisassemblyStartAddress;
		m_disassembler->FillDisassemblyListView(m_sourceListView, m_sourceDisassemblyStartAddress, m_currentInstructionAddress, m_currentStackFrameAddress, %m_sourceBreakpoints, %m_addressToSource, %m_sourceStartAddress, m_ignoreSource);
	}
}

void MainForm::RefreshCallStack()
{
	m_callStackListView->BeginUpdate();
	m_callStackListView->Items->Clear();

	// add PC as a virtual stack frame
	m_stackFrames.Add(m_machineState->RegPC-1);

	m_stackFrames.Reverse();
	
	for (int frameIndex = 0; frameIndex < m_stackFrames.Count; frameIndex++)
	{
		unsigned int stackAddr = m_stackFrames.default[frameIndex];
		String^ calledRoutineName = frameIndex == m_stackFrames.Count - 1 ? "start" : "???";
		String^ currentBlockName;

		if (frameIndex < m_stackFrames.Count - 1)
		{
			// child stack frames take the name with which they were called using JSR
			unsigned int callerAddress = m_stackFrames.default[frameIndex+1];
			if (m_machineState->Peek(callerAddress) == 0x20) // JSR
			{
				unsigned int calledAddress = m_machineState->Peek(callerAddress+1) + m_machineState->Peek(callerAddress+2) * 256 + (callerAddress & 0xFF0000);
				FileAndLine newFileAndLine;
				if (m_addressToSource.TryGetValue(calledAddress, newFileAndLine) && newFileAndLine.sourceLabel)
				{
					calledRoutineName = newFileAndLine.sourceLabel;
				}
			}
			else if (m_machineState->Peek(callerAddress) == 0x22) // JSL
			{
				unsigned int calledAddress = m_machineState->Peek(callerAddress+1) + m_machineState->Peek(callerAddress+2) * 256 + m_machineState->Peek(callerAddress+3) * 65536;
				FileAndLine newFileAndLine;
				if (m_addressToSource.TryGetValue(calledAddress, newFileAndLine) && newFileAndLine.sourceLabel)
				{
					calledRoutineName = newFileAndLine.sourceLabel;
				}
			}

			currentBlockName = GetNameForStackFrame(stackAddr);
		}
		else
		{
			// the top-most stack frame takes its name from the closest label to the current PC
			currentBlockName = GetNameForStackFrame(stackAddr);
		}

		String^ frameName;
		if (calledRoutineName->Equals(currentBlockName))
		{
			frameName = currentBlockName;
		}
		else
		{
			frameName = String::Format("{0}:{1}", calledRoutineName, currentBlockName);
		}

		ListViewItem^ frameItem = gcnew ListViewItem(String::Format("{0:X6} {1}", stackAddr, frameName));
		frameItem->Tag = stackAddr;

		if (frameIndex == 0)
		{
			frameItem->ImageKey = "arrow";
		}
		else if (stackAddr == m_currentStackFrameAddress)
		{
			frameItem->ImageKey = "stackframe";
		}

		m_callStackListView->Items->Add(frameItem);
	}

	m_stackFrames.Reverse();

	// remove PC as virtual stack frame
	m_stackFrames.RemoveAt(m_stackFrames.Count-1);

	m_callStackListView->EndUpdate();
	//m_callStackListView->EnsureVisible(currentStackFrameIndex);
}

String^ MainForm::GetNameForStackFrame(unsigned long addr)
{
	int searchCount = 0;
	unsigned long testAddress = addr;
	while (searchCount++ < 4000) // may need to skip up to 4000 bytes
	{
		FileAndLine newFileAndLine;
		if (m_addressToSource.TryGetValue(testAddress, newFileAndLine))
		{
			if (newFileAndLine.sourceLabel)
			{
				return newFileAndLine.sourceLabel;
			}			
		}

		testAddress--;
	}		

	return "";
}

// Refresh the micro source disassembly view
void MainForm::RefreshMicroDisassemblyView()
{
	m_microDisassembler->FillDisassemblyListView(m_microSourceListView, %m_microSourceBreakpoints);
}

// Fill a memory view textbox
void MainForm::FillMemoryTextBox(TextBox^ textBox, unsigned long startAddress, unsigned long highlightAddress, bool highlight)
{
	const int linesToShow = 32;

	String^ result;

	for (int line=0; line<linesToShow; line++)
	{
		result += String::Format("{0:X6}", startAddress + line*16);
		for (int byte=0; byte<16; byte++)
		{
			if (highlight && startAddress + line*16 + byte == highlightAddress)
			{
				result += ">";
			}
			else
			{
				result += " ";
			}

			unsigned char data = m_machineState->Peek(startAddress + line*16 + byte);
			result += String::Format("{0:X2}", data);
		}

		if (line < linesToShow-1)
		{
			result += "\r\n";
		}
	}

	textBox->Text = result;
}


// Fill console textbox
void MainForm::FillConsoleTextBox(TextBox^ textBox)
{
	const int linesToShow = 30;
	unsigned long startAddress = 0x200000;

	String^ result;

	for (int line=0; line<linesToShow; line++)
	{
		for (int byte=0; byte<128; byte+=2)
		{
			unsigned char data = m_machineState->Peek(startAddress + (line*8)*128 + byte);
			result += Char::ConvertFromUtf32(data);
		}

		if (line < linesToShow-1)
		{
			result += "\r\n";
		}
	}

	textBox->Text = result;
}

// Handle a click on the micro-step button
void MainForm::m_microStepButton_Click(System::Object^  sender, System::EventArgs^  e) 
{
	m_machineState->MicroStep();

	if (m_machineState->Phase == 0)
	{
		// PC points to the next byte after the instruction
		m_currentInstructionAddress = m_machineState->RegPC - 1;
		RetireInstruction();
	}

	m_currentStackFrameAddress = m_currentInstructionAddress;

	RefreshData();
}

void MainForm::m_stepButton_Click(System::Object^  sender, System::EventArgs^  e)
{
	// For JSR and JSL, set a breakpoint at the next instruction, and go
	if (m_machineState->RegOpcode == 0x20 || m_machineState->RegOpcode == 0x22)
	{
		int opSize = m_disassembler->GetOpcodeSize(m_machineState->RegOpcode);
		if (opSize)
		{
			// set a breakpoint at the next instruction, then run
			m_stepOverDestinationAddress = m_machineState->RegPC + opSize - 1;
			m_consoleTextBox->Focus();
			m_consoleTextBox->SelectionLength = 0;
			m_continueButtton->Enabled = false;
			m_stepButton->Enabled = false;
			m_stepInButton->Enabled = false;
			m_stepOutButton->Enabled = false;
			m_pauseButton->Enabled = true;
			m_microStepButton->Enabled = false; // this causes the LCD text to become selected?
			m_clearBreakpointsButton->Enabled = false;
			
			m_running = true;
		}
	}
	else
	{
		m_stepInButton_Click(sender, e);
	}
}


void MainForm::m_stepOutButton_Click(System::Object^  sender, System::EventArgs^  e)
{
	m_steppingOut = true;
	m_stepOutCount = 0;
	m_consoleTextBox->Focus();
	m_consoleTextBox->SelectionLength = 0;
	m_continueButtton->Enabled = false;
	m_stepButton->Enabled = false;
	m_stepInButton->Enabled = false;
	m_stepOutButton->Enabled = false;
	m_microStepButton->Enabled = false;
	m_clearBreakpointsButton->Enabled = false;
	m_pauseButton->Enabled = true;

	

	m_running = true;
}

// Handle a click on the step button
void MainForm::m_stepInButton_Click(System::Object^  sender, System::EventArgs^  e)
{
	do
	{
		m_machineState->MicroStep();
	} while (m_machineState->Phase != 0);

	// PC points to the next byte after the instruction
	m_currentInstructionAddress = m_machineState->RegPC - 1;
	m_currentStackFrameAddress = m_currentInstructionAddress;

	RetireInstruction();

	RefreshData();
}

// Handle a click on the reset button
void MainForm::m_resetButton_Click(System::Object^  sender, System::EventArgs^  e)
{
	m_machineState->Reset();
	m_consoleTextBox->Focus();
	m_consoleTextBox->SelectionLength = 0;
	m_pauseButton->Enabled = false;
	m_continueButtton->Enabled = true;
	m_stepButton->Enabled = true;
	m_stepInButton->Enabled = true;
	m_stepOutButton->Enabled = true;
	m_microStepButton->Enabled = true;
	m_clearBreakpointsButton->Enabled = true;
	m_running = false;
	m_currentInstructionAddress = m_machineState->RegPC - 1;
	m_currentStackFrameAddress = m_currentInstructionAddress;
	m_stepOverDestinationAddress = 0;
	m_steppingOut = false;
	m_stepOutCount = 0;
	m_executingOpcode = 0xFF;
	
	m_stackFrames.Clear();

	RefreshData();
}

// Handle a click on the Go button
void MainForm::m_continueButtton_Click(System::Object^  sender, System::EventArgs^  e)
{
	m_pauseButton->Enabled = true;
	m_consoleTextBox->Focus();
	m_consoleTextBox->SelectionLength = 0;
	m_continueButtton->Enabled = false;
	m_stepButton->Enabled = false;
	m_stepInButton->Enabled = false;
	m_stepOutButton->Enabled = false;
	m_microStepButton->Enabled = false;
	m_clearBreakpointsButton->Enabled = false;
	
	

	m_running = true;
}

// Handle a click on the pause button
void MainForm::m_pauseButton_Click(System::Object^  sender, System::EventArgs^  e)
{
	m_continueButtton->Enabled = true;
	m_continueButtton->Focus();
	//m_consoleTextBox->Focus();
	m_pauseButton->Enabled = false;
	m_stepButton->Enabled = true;
	m_stepInButton->Enabled = true;
	m_stepOutButton->Enabled = true;
	m_microStepButton->Enabled = true;
	m_clearBreakpointsButton->Enabled = true;
	m_running = false;
	m_steppingOut = false;
	m_stepOutCount = 0;

	m_stepOverDestinationAddress = 0;
	
	RefreshData();
}

// Handle the "Open Program ROM" menu item
void MainForm::openProgramROMToolStripMenuItem_Click(System::Object^  sender, System::EventArgs^  e)
{
	OpenFileDialog^ dialog = gcnew OpenFileDialog();
	dialog->Title = "Select Program File";
	dialog->Filter = "bin files (*.bin)|*.bin";
	dialog->RestoreDirectory = true;

	if (dialog->ShowDialog() == System::Windows::Forms::DialogResult::OK)
	{
		m_bootImage = dialog->FileName;
		Application::UserAppDataRegistry->SetValue("Boot Image", m_bootImage);
		InitializeSimulation();
		m_machineState = gcnew MachineState(dialog->FileName, ".", m_loadAddress, m_bootAddress);
		RefreshData();
		RefreshDisassemblyView(true);
		//m_disassembler->FillDisassemblyListView(m_sourceListView, m_sourceDisassemblyStartAddress, m_currentInstructionAddress, %m_sourceBreakpoints, %m_stackFrames, %m_addressToSource, %m_sourceStartAddress, m_ignoreSource);
	}
}

// Handle the "Open Source Labels" menu item
void MainForm::toolStripMenuItem2_Click(System::Object^  sender, System::EventArgs^  e)
{
	OpenFileDialog^ dialog = gcnew OpenFileDialog();
	dialog->Title = "Select Symbol File";
	dialog->Filter = "symbol files (*.sym)|*.sym|label files (*.lbl)|*.lbl";
	dialog->RestoreDirectory = true;

	if (dialog->ShowDialog() == System::Windows::Forms::DialogResult::OK)
	{
		LoadSourceLabels(dialog->FileName);
	}
}

// Load the label data for source debugging
void MainForm::LoadSourceLabels(String^ filename)
{
	m_addressToSource.Clear();
	m_sourceStartAddress.Clear();

	// try lbl file if sym doesn't exist, or vice-versa
	if (!File::Exists(filename))
	{
		if (filename->EndsWith(".sym"))
		{
			filename = filename->Substring(0,filename->LastIndexOf(".")) + ".lbl";
		}
		else if (filename->EndsWith(".lbl"))
		{
			filename = filename->Substring(0,filename->LastIndexOf(".")) + ".sym";
		}
	}

	bool isCC65 = filename->EndsWith(".lbl");

	String^ dir = ".\\";

	int lastSlash = filename->LastIndexOf('\\');
	if (lastSlash >= 0)
	{
		dir = filename->Substring(0, lastSlash+1);
	}

	try
	{
		StreamReader^ sr = gcnew StreamReader(filename);

		try
		{
			String^ line;
		 
			// Read and display lines from the file until the end of 
			// the file is reached.
			while ( (line = sr->ReadLine()) != nullptr )
			{
				if (line->StartsWith("L_"))
				{
					int comment = line->IndexOf(';');
					line = line->Substring(0, comment);

					array<String^>^ parts = line->Split('=');
					String^ labelName = parts[0]->Substring(2, parts[0]->Length-3);
					unsigned long sourceAddress = UInt32::Parse(parts[1]->Substring(1), NumberStyles::HexNumber) + (m_loadAddress & 0xFF0000);

					int idx = labelName->LastIndexOf('_');
					String^ sourceFilename = labelName->Substring(0, idx);

					int sourceLineNumber = Int32::Parse(labelName->Substring(idx+1));

					idx = sourceFilename->LastIndexOf('_');
					sourceFilename = sourceFilename->Substring(0, idx) + "." + sourceFilename->Substring(idx+1);
					sourceFilename = dir + sourceFilename;

					// addresses should be mapped to the highest line number possible in a file,
					// but to the file with the lowest line number
					FileAndLine fileAndLine;
					if (m_addressToSource.TryGetValue(sourceAddress, fileAndLine))
					{
						if (!fileAndLine.filename)
						{
							fileAndLine.filename = sourceFilename;
							fileAndLine.lineNumber = sourceLineNumber;
							m_addressToSource[sourceAddress] = fileAndLine;
						}
						else if (fileAndLine.filename->Equals(sourceFilename))
						{
							// same file, choose highest line
							if (sourceLineNumber > fileAndLine.lineNumber)
							{
								fileAndLine.lineNumber = sourceLineNumber;
								m_addressToSource[sourceAddress] = fileAndLine;
							}
						}
						else
						{
							// different file, choose lowest line
							if (sourceLineNumber < fileAndLine.lineNumber)
							{
								fileAndLine.filename = sourceFilename;
								fileAndLine.lineNumber = sourceLineNumber;
								m_addressToSource[sourceAddress] = fileAndLine;
							}
						}
					}
					else
					{
						// insert new entry for this address
						fileAndLine.filename = sourceFilename;
						fileAndLine.lineNumber = sourceLineNumber;
						fileAndLine.sourceLabel = "";
						m_addressToSource.Add(sourceAddress, fileAndLine);
					}
					
					unsigned long sourceStartAddress;
					if (m_sourceStartAddress.TryGetValue(sourceFilename, sourceStartAddress))
					{
						if (sourceAddress < sourceStartAddress)
						{
							m_sourceStartAddress[sourceFilename] = sourceAddress;
						}
					}
					else
					{
						// insert a new entry for the start address
						m_sourceStartAddress[sourceFilename] = sourceAddress;
					}
				}

				else
				{
					FileAndLine fileAndLine;
					unsigned long sourceAddress;

					if (isCC65)
					{
						array<String^>^ parts = line->Split('.');
						sourceAddress = UInt32::Parse(parts[0]->Substring(3,6), NumberStyles::HexNumber) + (m_loadAddress & 0xFF0000);
								
						m_addressToSource.TryGetValue(sourceAddress, fileAndLine);
						
						// update entry
						fileAndLine.sourceLabel = parts[1]->Trim();
						fileAndLine.filename = "unknown";
						fileAndLine.lineNumber = 1;
						m_addressToSource[sourceAddress] = fileAndLine;
					}
					else
					{
						int comment = line->IndexOf(';');
						if (comment >= 0)
						{
							line = line->Substring(0, comment);
						}

						array<String^>^ parts = line->Split('=');
						sourceAddress = UInt32::Parse(parts[1]->Substring(1), NumberStyles::HexNumber) + (m_loadAddress & 0xFF0000);
						
						m_addressToSource.TryGetValue(sourceAddress, fileAndLine);
						
						// update entry
						fileAndLine.sourceLabel = parts[0]->Trim();
						m_addressToSource[sourceAddress] = fileAndLine;
					}

					// keyboard handler labels?
					if (fileAndLine.sourceLabel->Equals("KEYBOARD_ADR_KEY_BUFFER"))
					{
						m_keyboardAdrKeyBuffer = sourceAddress;
					}
					else if (fileAndLine.sourceLabel->Equals("KEYBOARD_ADR_MODIFIER_BUFFER"))
					{
						m_keyboardAdrModifierBuffer = sourceAddress;
					} 
					else if (fileAndLine.sourceLabel->Equals("KEYBOARD_ADR_KEY_READY"))
					{
						m_keyboardAdrKeyReady = sourceAddress;
					} 
				}
			}

			m_disassembler->FillDisassemblyListView(m_sourceListView, m_sourceDisassemblyStartAddress, m_currentInstructionAddress, m_currentStackFrameAddress, %m_sourceBreakpoints, %m_addressToSource, %m_sourceStartAddress, m_ignoreSource);			
		}
		finally
		{
			if ( sr )
			delete (IDisposable^)sr;	
		}
	}
	catch ( Exception^ ) 
	{
		MessageBox::Show( String::Format("Error opening symbol file {0}", filename), "File Error" );
	}
}

// Handle a keypress in the LCD or console text box
void MainForm::m_lcdTextBox_KeyPress(System::Object^  sender, System::Windows::Forms::KeyPressEventArgs^  e) 
{
	// emulate the behavior of the keyboard driver, but don't signal an interrupt
	unsigned char byteValue = static_cast<unsigned char>(e->KeyChar);
	m_machineState->Store(m_keyboardAdrKeyBuffer, byteValue);
	m_machineState->Store(m_keyboardAdrModifierBuffer, 0);
	m_machineState->Store(m_keyboardAdrKeyReady, 1);
	e->Handled = true;
}

void MainForm::m_ShowDisassemblyCheckbox_CheckedChanged(System::Object^  sender, System::EventArgs^  e) 
{
	m_ignoreSource = m_ShowDisassemblyCheckbox->Checked;
	m_disassembler->FillDisassemblyListView(m_sourceListView, m_sourceDisassemblyStartAddress, m_currentInstructionAddress, m_currentStackFrameAddress, %m_sourceBreakpoints, %m_addressToSource, %m_sourceStartAddress, m_ignoreSource);
}

void MainForm::toolStripMenuItem3_Click(System::Object^  sender, System::EventArgs^  e)
{
	InputDialog^ inputDialog = gcnew InputDialog();
	inputDialog->Text = "Set Program Load Address";
	inputDialog->SetPrompt("Load Address (hex)");
	inputDialog->SetInputBoxText(String::Format("{0:X6}", m_loadAddress));

	System::Windows::Forms::DialogResult result = inputDialog->ShowDialog();
	if (result == System::Windows::Forms::DialogResult::OK)
	{
		m_loadAddress = System::Convert::ToUInt32(inputDialog->GetInputBoxText(), 16);
		Application::UserAppDataRegistry->SetValue("Load Address", m_loadAddress);
	}

}

void MainForm::toolStripMenuItem4_Click(System::Object^  sender, System::EventArgs^  e)
{
	InputDialog^ inputDialog = gcnew InputDialog();
	inputDialog->Text = "Set Program Start Address";
	inputDialog->SetPrompt("Start Address (hex)");
	inputDialog->SetInputBoxText(String::Format("{0:X6}", m_bootAddress));

	System::Windows::Forms::DialogResult result = inputDialog->ShowDialog();
	if (result == System::Windows::Forms::DialogResult::OK)
	{
		m_bootAddress = System::Convert::ToUInt32(inputDialog->GetInputBoxText(), 16);
		Application::UserAppDataRegistry->SetValue("Boot Address", m_bootAddress);
	}
}

void MainForm::toolStripMenuItem5_Click(System::Object^  sender, System::EventArgs^  e)
{
InputDialog^ inputDialog = gcnew InputDialog();
	inputDialog->Text = "Set Memory Breakpoint";
	inputDialog->SetPrompt("Breakpoint Address (hex)");
	inputDialog->SetInputBoxText(String::Format("{0:X6}", m_memoryBreakpointAddress));

	System::Windows::Forms::DialogResult result = inputDialog->ShowDialog();
	if (result == System::Windows::Forms::DialogResult::OK)
	{
		m_memoryBreakpointAddress = System::Convert::ToUInt32(inputDialog->GetInputBoxText(), 16);
	}
}

void MainForm::toolStripMenuItem6_Click(System::Object^  sender, System::EventArgs^  e)
{
	m_memoryBreakpointAddress = 0xFFFFFF;
}