#include "stdafx.h"
#include "MicroDisassembly.h"
#include "MachineState.h"

using namespace Simulator;
using namespace System::Drawing;
using namespace System::Windows::Forms;

MicroDisassembly::MicroDisassembly(MachineState^ machineState)
{	
	m_machineState = machineState;
}

void MicroDisassembly::FillDisassemblyTextBox(TextBox^ textBox)
{
	String^ result;
	unsigned char currentPhase = m_machineState->Phase;
	unsigned long currentMicroAddress = m_machineState->GetCurrentMicroaddress();

	for (unsigned char phase=0; phase<16; phase++)
	{
		m_microInstructionLookup.Clear();
		
		for (unsigned char a=0; a<2; a++)
		{
			for (unsigned char c=0; c<2; c++)
			{
				for (unsigned char n=0; n<2; n++)
				{
					for (unsigned char z=0; z<2; z++)
					{
						for (unsigned char v=0; v<2; v++)
						{
							unsigned long microOp = m_machineState->GetMicroinstruction(m_machineState->RegOpcode * 512 + phase * 32 + v*16 + z*8 + n*4 + c*2 + a);
							
							if (!m_microInstructionLookup.ContainsKey(microOp))
							{
								m_microInstructionLookup[microOp] = MicroInstructionData();
							}

							MicroInstructionData data = m_microInstructionLookup[microOp];
							data.AddressCarry |= 1<<a;
							data.Carry |= 1<<c;
							data.Negative |= 1<<n;
							data.Zero |= 1<<z;
							data.Overflow |= 1<<v;
							m_microInstructionLookup[microOp] = data;
						}
					}
				}
			}
		}

		bool lastPhase = true;

		for each (KeyValuePair<unsigned long,MicroInstructionData> entry in m_microInstructionLookup)
		{
			String^ conditionString = "";

			conditionString += ConditionString(entry.Value.Overflow, "V", true);
			conditionString += ConditionString(entry.Value.Zero, "Z", true);
			conditionString += ConditionString(entry.Value.Negative, "N", false);
			conditionString += ConditionString(entry.Value.Carry, "C", true);
			conditionString += ConditionString(entry.Value.AddressCarry, "A", true);

			// strip trailing ampersand
			if (conditionString->EndsWith("&"))
			{
				conditionString = "if (" + 
					conditionString->Substring(0, conditionString->Length-1) + ") ";

				if (conditionString->Length % 2 == 1)
				{
					conditionString += " ";
				}
			}

			String^ operationString = MicroOpDescription(entry.Key);
			
			// strip out do-nothing data move if there's anything else being done
			if (operationString->StartsWith("A <- A; "))
			{
				operationString = operationString->Substring(8);
			}

			if (phase == currentPhase &&
				(entry.Value.AddressCarry & (1 << (currentMicroAddress & 0x1))) &&
				(entry.Value.Carry & (1 << ((currentMicroAddress >> 1) & 0x1))) &&
				(entry.Value.Negative & (1 << ((currentMicroAddress >> 2) & 0x1))) &&
				(entry.Value.Zero & (1 << ((currentMicroAddress >> 3) & 0x1))) &&
				(entry.Value.Overflow & (1 << ((currentMicroAddress >> 4) & 0x1))))
			{
				result += String::Format(">{0:X1} {1}{2}\r\n", phase, conditionString, operationString);
			}
			else
			{
				result += String::Format(" {0:X1} {1}{2}\r\n", phase, conditionString, operationString);
			}

			if (MachineState::IsBitOne(entry.Key, MachineState::BitPos::kLoadOp))
			{
				lastPhase = false;
			}
		}

		// if all CC's do a nextOp on this phase, then don't list the remaining phases
		if (lastPhase)
		{
			break;
		}
	}

	textBox->Text = result;
}

String^ MicroDisassembly::ConditionString(unsigned char bits, String^ name, bool isActiveLow)
{
	if (bits == 1 && !isActiveLow ||
		bits == 2 && isActiveLow)
	{
		return "!" + name + "&";
	} 
	else if (bits == 2 && !isActiveLow ||
			 bits == 1 && isActiveLow)
	{
		return name + "&";
	}
	else
	{
		return "";
	}

}

String^ MicroDisassembly::DestinationString(unsigned long microOp)
{
	unsigned long dataDest = MachineState::FieldValue(microOp, MachineState::BitPos::kDest, 3);
	if (MachineState::IsBitOne(microOp, MachineState::BitPos::kDest3))
	{
		dataDest += 8;
	}
	// determine the move destination
	if ((unsigned long)MachineState::Dest::kLoadSPHI == dataDest)
	{
		return "SPHI";
	}
	else if ((unsigned long)MachineState::Dest::kLoadSPBank == dataDest)
	{
		return "SPBANK";
	}
	else if ((unsigned long)MachineState::Dest::kLoadARLO == dataDest)
	{
		return "ARLO";
	}
	else if ((unsigned long)MachineState::Dest::kLoadARHI == dataDest)
	{
		return "ARHI";
	}
	else if ((unsigned long)MachineState::Dest::kLoadARBank == dataDest)
	{
		return "ARBANK";
	}
	else if ((unsigned long)MachineState::Dest::kLoadPCLO == dataDest)
	{
		return "PCLO";
	}
	else if ((unsigned long)MachineState::Dest::kLoadPCHI == dataDest)
	{
		return "PCHI";
	}
	else if ((unsigned long)MachineState::Dest::kLoadPCBank == dataDest)
	{
		return "PCBANK";
	}
	else if ((unsigned long)MachineState::Dest::kLoadMem == dataDest)
	{
		unsigned long addressSelect = MachineState::FieldValue(microOp, MachineState::BitPos::kAddressSelect, 2);
		if ((unsigned long)MachineState::AddressSelect::kSelectSP == addressSelect)
		{
			return "MEM(SP)";
		}
		else if ((unsigned long)MachineState::AddressSelect::kSelectAR == addressSelect)
		{
			return "MEM(AR)";
		}
		else
		{
			return "MEM(PC)";
		}
	}
	else if ((unsigned long)MachineState::Dest::kLoadSPLO == dataDest)
	{
		return "SPLO";
	}
	else if ((unsigned long)MachineState::Dest::kLoadRegT == dataDest)
	{
		return "T";
	}
	else if ((unsigned long)MachineState::Dest::kLoadRegY == dataDest)
	{
		return "Y";
	}
	else if ((unsigned long)MachineState::Dest::kLoadRegX == dataDest)
	{
		return "X";
	}
	else if ((unsigned long)MachineState::Dest::kLoadRegA == dataDest)
	{
		return "A";
	}
	else
	{
		return "?";
	}
}

String^ MicroDisassembly::LeftALUString(unsigned long microOp)
{
	unsigned long leftOp = MachineState::FieldValue(microOp, MachineState::BitPos::kALULeft, 2);
	if ((unsigned long)MachineState::ALULeft::kLeftZero == leftOp)
	{
		return "0";
	}
	else if ((unsigned long)MachineState::ALULeft::kLeftRegX == leftOp)
	{
		return "X";
	}
	else if ((unsigned long)MachineState::ALULeft::kLeftRegA == leftOp)
	{
		return "A";
	}
	else if ((unsigned long)MachineState::ALULeft::kLeftRegYorX7 == leftOp)
	{
		// y and x7 share a control bit: use x7 with address register destinations, otherwise use Y
		unsigned long dest3 = MachineState::FieldValue(microOp, MachineState::BitPos::kDest3, 1);
		if (dest3)
		{
			return "Y";
		}
		else
		{
			return "X7";
		}
	}
	else
	{
		return "?";
	}
}

String^ MicroDisassembly::RightALUString(unsigned long microOp)
{
	unsigned long rightOp = MachineState::FieldValue(microOp, MachineState::BitPos::kALURight, 2);
	if ((unsigned long)MachineState::ALURight::kRightAddressLo == rightOp ||
		(unsigned long)MachineState::ALURight::kRightAddressHi == rightOp ||
		(unsigned long)MachineState::ALURight::kRightAddressBank == rightOp)
	{
		unsigned long addressSelect = MachineState::FieldValue(microOp, MachineState::BitPos::kAddressSelect, 2);
		if ((unsigned long)MachineState::AddressSelect::kSelectSP == addressSelect)
		{
			return "SP" + AddressByteName(rightOp);
		}
		else if ((unsigned long)MachineState::AddressSelect::kSelectAR == addressSelect)
		{
			return "AR" + AddressByteName(rightOp);
		}
		else
		{
			return "PC" + AddressByteName(rightOp);
		}
	}
	else if ((unsigned long)MachineState::ALURight::kRightRegT == rightOp)
	{
		return "T";
	}
	else
	{
		return "?";
	}
}

String^ MicroDisassembly::AddressByteName(unsigned long rightOp)
{
	if ((unsigned long)MachineState::ALURight::kRightAddressLo == rightOp)
	{
		return "LO";
	}
	else if ((unsigned long)MachineState::ALURight::kRightAddressHi == rightOp)
	{
		return "HI";
	}
	else
	{
		return "BANK";
	}
}

String^ MicroDisassembly::MicroOpDescription(unsigned long microOp)
{
	String^ result = DestinationString(microOp) + " <- ";

	if (MachineState::IsBitZero(microOp, MachineState::BitPos::kDriveALU))
	{
		unsigned char c = (unsigned char)MachineState::FieldValue(microOp, MachineState::BitPos::kALUCarry, 1);
		unsigned long function = MachineState::FieldValue(microOp, MachineState::BitPos::kALUOp, 4);
		
		if (MachineState::IsBitOne(microOp, MachineState::BitPos::kALUMode))
		{
			// logical
			switch (function)
			{
				case 0:
					result += "~" +  LeftALUString(microOp);
					break;
				case 3:
					result += "0";
					break;
				case 5:
					result += "~" + RightALUString(microOp);
					break;
				case 6:
					result += LeftALUString(microOp) + "^" + RightALUString(microOp);
					break;
				case 10:
					result += RightALUString(microOp);
					break;
				case 11:
					result += LeftALUString(microOp) + "&" + RightALUString(microOp);
					break;
				case 12:
					result += "$FF";
					break;
				case 14:
					result += LeftALUString(microOp) + "|" + RightALUString(microOp);
					break;
				case 15:
					result += LeftALUString(microOp);
					break;
				default:
					result += "ALU(" + LeftALUString(microOp) + "," + RightALUString(microOp) + "," + String::Format("${0:X1},{1},1", function, c) + ")";
					break;
			}
		}
		else
		{
			// arithmetic
			switch (function)
			{
				case 0:
					result += c ? LeftALUString(microOp) : LeftALUString(microOp) + "+1";
					break;
				case 3:
					result += c ? "-1" : "0";
					break;
				case 6:
					result += c ? LeftALUString(microOp) + "-" + RightALUString(microOp) + "-1": LeftALUString(microOp) + "-" + RightALUString(microOp);
					break;
				case 9:
					result += c ? LeftALUString(microOp) + "+" + RightALUString(microOp): LeftALUString(microOp) + "+" + RightALUString(microOp) + "+1";
					break;
				case 12:
					result += c ? LeftALUString(microOp) + "+" + LeftALUString(microOp) : LeftALUString(microOp) + "+" + LeftALUString(microOp) + "+1";
					break;
				case 15:
					result += c ? LeftALUString(microOp) + "-1" : LeftALUString(microOp);
					break;
				default:
					result += "ALU(" + LeftALUString(microOp) + "," + RightALUString(microOp) + "," + String::Format("${0:X1},{1},0", function, c) + ")";
					break;
			}
		}
	}
	else
	{
		unsigned long addressSelect = MachineState::FieldValue(microOp, MachineState::BitPos::kAddressSelect, 2);
		if ((unsigned long)MachineState::AddressSelect::kSelectSP == addressSelect)
		{
			result += "MEM(SP)";
		}
		else if ((unsigned long)MachineState::AddressSelect::kSelectAR == addressSelect)
		{
			result += "MEM(AR)";
		}
		else
		{
			result += "MEM(PC)";
		}
	}

	// condition codes
	unsigned long s0 = MachineState::FieldValue(microOp, MachineState::BitPos::kS0, 1);
	unsigned long s1 = MachineState::FieldValue(microOp, MachineState::BitPos::kS1, 1);
	if (s1 && s0)
	{
		if (DestinationString(microOp)->Equals("A") &&
			RightALUString(microOp)->Equals("T"))
		{
			result += "; latchAllCC";
		}
		else
		{
			result += "; latchNZC";
		}
	}
	else if (s1 && !s0)
	{
		result += "; shiftCC";
	}
	else if (!s1 && s0)
	{
		result += "; latchNZ";
	}

	// opcode
	if (MachineState::IsBitZero(microOp, MachineState::BitPos::kLoadOp))
	{
		unsigned long addressSelect = MachineState::FieldValue(microOp, MachineState::BitPos::kAddressSelect, 2);
		if ((unsigned long)MachineState::AddressSelect::kSelectSP == addressSelect)
		{
			result += "; OP <- MEM(SP)";
		}
		else if ((unsigned long)MachineState::AddressSelect::kSelectAR == addressSelect)
		{
			result += "; OP <- MEM(AR)";
		}
		else
		{
			result += "; OP <- MEM(PC)";
		}
	}

	// special
	unsigned long special = MachineState::FieldValue(microOp, MachineState::BitPos::kSpecial, 2);
	if ((unsigned long)MachineState::Special::kLoadIE == special)
	{
		if (MachineState::IsBitZero(microOp, MachineState::BitPos::kSPDirIEData))
		{
			result += "; enableInt";
		}
		else
		{
			result += "; disableInt";
		}
	}
	else if ((unsigned long)MachineState::Special::kIncAR == special)
	{
		result += "; AR++";
	}
	else if ((unsigned long)MachineState::Special::kIncPC == special)
	{
		result += "; PC++";
	}

	// stack pointer
	if (MachineState::IsBitZero(microOp, MachineState::BitPos::kCountSP))
	{
		if (MachineState::IsBitOne(microOp, MachineState::BitPos::kSPDirIEData))
		{
			result += "; SP++";
		}
		else
		{
			result += "; SP--";
		}
	}

	return result;
}

void MicroDisassembly::FillDisassemblyListView(ListView^ listView, Dictionary<unsigned long, List<unsigned long>^>^ microSourceBreakpoints)
{
	listView->BeginUpdate();
	listView->Items->Clear();

	unsigned char currentPhase = m_machineState->Phase;
	unsigned long currentMicroAddress = m_machineState->GetCurrentMicroaddress();

	for (unsigned char phase=0; phase<16; phase++)
	{
		m_microInstructionLookup.Clear();
		
		for (unsigned char a=0; a<2; a++)
		{
			for (unsigned char c=0; c<2; c++)
			{
				for (unsigned char n=0; n<2; n++)
				{
					for (unsigned char z=0; z<2; z++)
					{
						for (unsigned char v=0; v<2; v++)
						{
							unsigned long microOp = m_machineState->GetMicroinstruction(m_machineState->RegOpcode * 512 + phase * 32 + v*16 + z*8 + n*4 + c*2 + a);
							
							if (!m_microInstructionLookup.ContainsKey(microOp))
							{
								m_microInstructionLookup[microOp] = MicroInstructionData();
							}

							MicroInstructionData data = m_microInstructionLookup[microOp];
							data.AddressCarry |= 1<<a;
							data.Carry |= 1<<c;
							data.Negative |= 1<<n;
							data.Zero |= 1<<z;
							data.Overflow |= 1<<v;
							m_microInstructionLookup[microOp] = data;
						}
					}
				}
			}
		}

		bool lastPhase = true;

		for each (KeyValuePair<unsigned long,MicroInstructionData> entry in m_microInstructionLookup)
		{
			ListViewItem^ lineItem = gcnew ListViewItem();
			unsigned long microProgramBaseAddress = (m_machineState->RegOpcode << 4) | phase;
			unsigned long microProgramConditionCodes = 
				(entry.Value.AddressCarry << 0) | (entry.Value.Carry << 2) | (entry.Value.Negative << 4) | (entry.Value.Zero << 6) | (entry.Value.Overflow << 8);
		
			lineItem->Tag = gcnew KeyValuePair<unsigned long, unsigned long>(microProgramBaseAddress, microProgramConditionCodes);
			lineItem->BackColor = phase % 2 ? Color::FromArgb(240, 240, 240) : Color::White;

			String^ conditionString = "";

			conditionString += ConditionString(entry.Value.Overflow, "V", true);
			conditionString += ConditionString(entry.Value.Zero, "Z", true);
			conditionString += ConditionString(entry.Value.Negative, "N", false);
			conditionString += ConditionString(entry.Value.Carry, "C", true);
			conditionString += ConditionString(entry.Value.AddressCarry, "A", true);

			// strip trailing ampersand
			if (conditionString->EndsWith("&"))
			{
				conditionString = "if (" + 
					conditionString->Substring(0, conditionString->Length-1) + ") ";

				if (conditionString->Length % 2 == 1)
				{
					conditionString += " ";
				}
			}

			String^ operationString = MicroOpDescription(entry.Key);
			
			// strip out do-nothing data move if there's anything else being done
			if (operationString->StartsWith("A <- A; "))
			{
				operationString = operationString->Substring(8);
			}

			lineItem->Text = String::Format("{0:X1} {1}{2}", phase, conditionString, operationString);

			if (phase == currentPhase &&
				(entry.Value.AddressCarry & (1 << (currentMicroAddress & 0x1))) &&
				(entry.Value.Carry & (1 << ((currentMicroAddress >> 1) & 0x1))) &&
				(entry.Value.Negative & (1 << ((currentMicroAddress >> 2) & 0x1))) &&
				(entry.Value.Zero & (1 << ((currentMicroAddress >> 3) & 0x1))) &&
				(entry.Value.Overflow & (1 << ((currentMicroAddress >> 4) & 0x1))))
			{
				lineItem->ImageKey = "arrow";

				if (microSourceBreakpoints->ContainsKey(microProgramBaseAddress))
				{
					List<unsigned long>^ conditionCodes = microSourceBreakpoints[microProgramBaseAddress];
					for each (unsigned long conditionCode in conditionCodes)
					{
						if (conditionCode == microProgramConditionCodes)
						{
							lineItem->ImageKey = "breakpoint with arrow";
							break;
						}
					}
				}
			}
			else if (microSourceBreakpoints->ContainsKey(microProgramBaseAddress))
			{
				List<unsigned long>^ conditionCodes = microSourceBreakpoints[microProgramBaseAddress];
				for each (unsigned long conditionCode in conditionCodes)
				{
					if (conditionCode == microProgramConditionCodes)
					{
						lineItem->ImageKey = "breakpoint";
						break;
					}
				}
			}

			if (MachineState::IsBitOne(entry.Key, MachineState::BitPos::kLoadOp))
			{
				lastPhase = false;
			}

			listView->Items->Add(lineItem);
		}

		// if all CC's do a nextOp on this phase, then don't list the remaining phases
		if (lastPhase)
		{
			break;
		}
	}

	listView->EndUpdate();
}